/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! \file Data.h
    \brief Contains the Data, List, and Grid classes.

    Data contains basic functions for manipulating the map of Attributes it contains.  List and Grid both specialize the Data object further and make use of their known dimensionality to map the data more efficiently among processes.
*/

#ifndef BASIN_DATA_H
#define BASIN_DATA_H
#include "Geometry.h"
#include "Attribute.h"

class Graphics_List;
class Graphics_Grid;
class Region;
class List;
class Grid;




//=====================================================

/** \class Data Data.h "Data.h"
    \brief Data is the parent class for both List and Grid.
    
    It consists of a map of strings to \link Attribute Attributes \endlink.  A Data also has an associated Geometry, including dimensionality and metric, and is associated within a Region.  Note that the addition and setting of attributes is handled by the derived Grid and List types, since those carry the necessary
nx or nElements information to properly define the Attribute.
*/
class Data {
 public:
  /** @name Data Constructors
      Functions to construct Data objects. */
  //@{
  // Empty Constructor Class
  Data();
  // Destructor
  virtual ~Data();
  
  // Constructor used to set the Data object's Geometry based on the Parent Region
  Data(Region* parent);
  //@}
  
  /** @name Data Properties
      Get dimensionality or contained attributes. */
  //@{
  //! Return the Dimensionality of the Geometry
  int get_ndims() {return myGeometry.get_ndims();};
  

  int get_num_attributes () {
	  return attributes.size ();
  }
  // Return a vector of strings listing attributes (mass, density, etc.)
  vector <string> get_properties();
  // Set the geometry associated with the data object
  int set_geometry(Geometry* geo)
    {
      myGeometry = Geometry(*geo);
      return 0;
    }
  //! Set the parent Region to be used
  int set_parent(Region* par)
    {
      parentRegion = par;
      return 0;
    }
  //!
  Region* get_parent(){return parentRegion;}


  int set_context(Context& c)
    {
      DataContext=c;
      return 0;
    }


  int set_context(vector<double>& vll, vector<double>& vur, vector<int> &vnxyz)
    {
      this->DataContext=Context(vll,vur,vnxyz,this);
      return 0;
    }

   Context* get_context_pointer() 
    {
      return &(this->DataContext);
    }


  /// Obtains the range of "rows" associated with a process
  /** get_distribution will return the rows of data in local memory of a process.  This is merely a division of the first dimension with the current method of distributing data.  iLo and iHi are the low and hi row numbers, the return is for error checking.  Use is deprecated.  Recommended to use Data::get_distribution_index instead.*/
  int get_distribution(string key,int procRank,int* iLo, int* iHi)
    {
      if(attributes.find(key)!=attributes.end()){
	attributes[key].get_gen_distribution(procRank,iLo,iHi);
	return 0;
      }
      else
	return -1;
    }
  /** get_distribution will return the indices of data in local memory of a process.  iLo and iHi are the low and hi index numbers, the return is for error checking.  An example usage is when we want each processor to work on a subset of the data and then pass it on essentially, as seen in List::compute_density.  An example:\n
\code
int low, hi;
int numProcs;
MPI_Comm_size(MPI_COMM_WORLD,&numProcs);
string myKey="DMDensity";
int myRank     = MPI::COMM_WORLD.Get_rank();
for(int i = myRank; i < myRank+numProcs; i++){
   if(get_distribution_index(myKey, i%numProcs, &low, &hi)==-1){
      cerr<< "Error: Could not obtain distribution of " << key << ".  Check if attribute exists.\n";
      return -1;
   }
   else {
      //now can properly size data array
      double* theData = new double[hi-lo+1];
      if(get_attribute_local(myKey, theData)==0)
         //some useful code using this processor's data
      else
         cerr ...
   }
}
\endcode

  */
  // Obtains the range of indices associated with a process
  int get_distribution_index(string key,int procRank,int* iLo, int* iHi)
    {
      if(attributes.find(key)!=attributes.end()){
	attributes[key].get_gen_distribution_index(procRank,iLo,iHi);
	return 0;
      }
      else
	return -1;
    }
  
  //@}

  /** @name Functions to Get Attributes
      Functions to get elements or entire attributes. They return 0 on success and -1 on a failure. With the exception of get_attribute_local, they all return copies of data.*/
  //@{
  
  /** You must pass an object of type T (corresponding to the Attribute
   * type), the string key identifying the particular attribute, and
   * the index (position inside the array.)
   */
  // Return (via the second argument) the value of an element in a Attribute.
  template <class T> int get_attribute_element(string key, int idx,T& outVal) {
    if(attributes.find(key)!=attributes.end())
      return attributes[key].get_element(idx,outVal);
    else
      return -1;
  }
  /** This version requires a pointer to an object of type T instead of the actually object.*/
  // Another version
  template<class T> int get_attribute_element(string key, int idx,T* &outVal) {
    if(attributes.find(key)!=attributes.end())
      return attributes[key].get_element( idx,outVal);
    else
      return -1;
  }
  
  ///Also can pass a vector
  template<class T> int get_attribute(string key, int iLo, int iHi, vector<T>* &outVal){
    if(attributes.find(key)!=attributes.end())
      return attributes[key].get(iLo,iHi,outVal);
    else
      return -1;
  }

  // returning an Attribute. To be used for deep copies of Attributes
  Attribute copy_attribute(string key ){ 
    if(attributes.find(key) !=attributes.end())
      return (attributes[key]);
    else
      return Attribute();
  }


  //!
  Attribute* get_attribute_pointer(string key ){ 
    if(attributes.find(key) !=attributes.end())
      {
	return &(attributes[key]);
      }
    else
      return new Attribute();
  }

  //! Return an attribute pointer based on the position of the iterator
  Attribute* get_attribute_pointer(int i) {
     map<string, Attribute>::iterator iter;
     int j=0;
     if (i < attributes.size()) {
       for(iter = attributes.begin(); iter != attributes.end(); iter++) {
	 if (i == j) {
	   string key=iter->first;
	   return &(attributes[key]);
	 }
	 j++;
       }
     }
     else return NULL;
  }

  //!
  int attribute_exist(string key) {
    if(attributes.find(key) !=attributes.end()) {
      return 1;
    } else {
      return 0;
    }
  }



  //@}

  
  /** @name Other Attribute Functions
      Non-accessor functions. */
  //@{
  //! Delete a particular attribute in a list/grid or other data object
  int delete_attribute(string key);

  // Not a great algorithm, but will change the name of a key in the map, i.e. "rename"
  bool change_attribute_key (string oldkey, string newkey);
  // A function formatted mostly for the gui client, returns newline delimitted string of attribute names
  string get_attribute_key_names (void);
  
  //! How many elements are in attribute?
  int get_attribute_size(string key)
    {
      if(attributes.find(key)!=attributes.end())
	return attributes[key].size();
      else
	return -1;
    }
  //@}


  string get_client_information (string info = "", string depth = "");


 protected:

  // Associated Geometry Object
  Geometry myGeometry;

  // Map of strings to \link Attribute Attributes\endlink that store actually store data elements
  map < string , Attribute > attributes;
  
  //Identify what Region contains this data member
  Region* parentRegion;

  //Context to be provided to Attributes
  Context DataContext;

  //These are:
  //ll=lower left corner
  //ur=upper right corner
  //nx=number of grid cells in each direction. For a list, nx[0]=nElements.
  //the size of each of these vectors must match
  //these are currently also in Geometry but we will eventually get rid of Geometry
  vector<double> ll;
  vector<double> ur;
  vector<int> nx;

};


// \a List Derived Class

/** \class List Data.h "Data.h"
    \brief List is a specific Data which is one-dimensional and has a "primary sort".

    The dimension of the primary sort is determined by the Geometry ("positions" within that space), and a primary sort parameter.  Currently it has basic statistical as well as density computation functions defined on it.*/


class List: public Data {

 public:
  
  /** @name List Constructors
      Functions to construct List objects, must be called by ALL nodes. */
  //@{
  //! Empty List Constructor
  List();

  /// List Constructor instantiates parentRegion by calling Data(Region*), 
  /// declares nElements which is size of the List.\n
  /// Must make sure to set nElements before populating the list.
  List(Region* parent, int nElementsNew = 0);
  //@}



  /** @name Functions to Add Attributes
      Functions to add attributes and set the value of them.  They return 0 on sucess and -1 on a failure.  ALL processors must call these functions to create the attribute.  newData is purely a dummy value to give the created Attribute type information.*/
  //@{
  ///Add an attribute
  template<class T> int add_attribute(string key,T* newData) {
    attributes[key]=Attribute(nElements,newData[0],key,&(this->DataContext));
    return 0;
  }
  /// Other version of add_attribute
  template<class T> int add_attribute(string key,T newData){
    //SOME PROBLEMS WITH THE FOLLOWING LINE 
    //PROBABLY RELATED TO THE COPY CONSTRUCTOR
    //    attributes[key]=Attribute(nElements,newData);
    //THE FOLLOWING LINE IS EQUIVALENT TO THE PREVIOUS ONE BUT MORE EFFICIENT 
    //and IT AVOID SOME constructor or = operator which are 
    //probably to be debugged
    //see Meyers book STL page 106.
    attributes.insert(map<string,Attribute>::value_type(key,Attribute(nElements,newData,key,(&(this->DataContext)))));
    return 0;
  }
  //! Add a whole Attribute
  int add_attribute(string key, Attribute& newVect) {
    //    attributes[key] = newVect;
    newVect.set_name(key);
    /* 
       This condition is supposed to check whether the List already has its own 
       non-default Context. If not the List is going to take it from the first 
       Attribute added to it if the Attribute has its own Context.
       If the Attribute has not its own Context, the List default Context is kept 
       until the first Attribute with a Context is added.
    */
    if (this->DataContext.get_n_grid_points().size()==1 && 
	this->DataContext.get_n_grid_points()[0]==2 && 
	newVect.get_context_pointer()->get_n_grid_points().size()!=1)
      {
	//getting the Context from the Attribute
        this->set_context(*(newVect.get_context_pointer()));
      }
    else
      {
	//providing the Attribute with a Context
	newVect.set_context(&(this->DataContext));
      }
    attributes.insert(map<string,Attribute>::value_type(key,newVect));
    return 0;
  }

  //@}

  /** @name Function to Add Data
      set_attribute puts the actual data into the attribute.
      Warning: There is no safety checks to prevent collisions.
      Be careful to not have more than one process try and set_attribute to the same range of of the attribute at the same time.
  */

  //@{

  ///Put data to the entire range of the attribute
  template<class T> int set_attribute(string key, T* newData) {
    if(attributes.find(key)!=attributes.end()){
      int temp = attributes[key].put(0,nElements-1,newData);
      //reset_selection();
      return temp;
    }
    else
      return -1;
  }
  /// Another version to put to the entire range
  template<class T> int set_attribute(string key, T newData) {
    return set_attribute(key, &newData);
  }
  /// Can pass it a vector too (entire range)
  template<class T> int set_attribute(string key, vector<T> &newData) {
    if(attributes.find(key)!=attributes.end()){
      int temp = attributes[key].put(0,nElements-1,newData);
      reset_selection();
      return temp;
    }
    else
      return -1;
  }

  /// set_attribute taking hi and low indices of the range
  template<class T> int set_attribute(string key, int iLo, int iHi, T* newData){
    if(attributes.find(key)!=attributes.end()){
      int temp=attributes[key].put(iLo,iHi,newData);
      reset_selection();
      return temp;
    }
    else
      return -1;
  }
  /// Another version using hi and low points
  template<class T> int set_attribute(string key, int iLo, int iHi, T newData){
    return set_attribute(key, iLo, iHi, &newData);
  }

  /// Change/replace an element of an attribute.   *Don't* update
  /// the selection list here (user should do this afterwards).
  /// Don't set multiple elements by repeated calls, use set_element with end indices
  template<class T> int set_attribute_element(string key, const T& newData,
					      int idx)
    {
      if(attributes.find(key)!=attributes.end())
	return attributes[key].set_element(idx,newData);
      else
	return -1;
    }
  //@}

  /** @name nElements Functions
      Get and set the nElements member variable.*/
  //@{
  //! How many elements?
  int get_nelements()	{return nElements;}

  //! Set (and check) nElements from the attributes.
  int set_nelements();
  //@}


  /** @name Selection Functions
      Functions dealing with selection and selectionHistory.*/
  //@{
  /// Reset the selection list and history.
  int reset_selection();

  /// Print the selection list.
  int print_selection(ostream  &s = cout);

  /// Print the selection history.
  int print_selection_history(ostream &s = cout);  

  /// Return a pointer to the selection list.
  list<int>* get_selection()	{return &selection;}

  /// Sort the list based on which attribute
  void sort_selection(string);

  //@}

  /** @name Position Functions
      Functions dealing with the position key and going from index<->position.  Also basic nearest calculations.*/
  //@{
  /// Sets attribute specified by key to become the special position key within the List (posKey)
  /** Note: Right now, there is NO checking to assure that this is
   * actually an ndims-dimensional vector of reals
   */
  int set_pos_key(string key) {
    if(attributes.find(key)!=attributes.end()){ posKey=key;return 0;}
    else{ cerr<<"Position Key not an attribute."<<endl; return -1;}
  }

  // get_pos() returns an ordered vector based on the scalar input index
  vector<double> get_pos(int idx);

  /// Return the index of the particle nearest a given point.
  int calculate_idx_nearest(const vector<double>& x0);

  /// Calculate the nth nearest particle with respect to a given particle
  /** You must send the particle depth and the index of the given particle*/
  int calculate_nth_nearest(int particleDepth, const vector<double>& testPos);

  //@}


  /** @name List Statistics Functions
      Computing basic statistical values and histograms.*/
  //@{
  //Compute basic statistics on an attribute
  int compute_statistics(string key, string weight="", ostream& s=cout);
  /**< This computes the mean, standard deviation, and 10-percentiles on the attribute specified by key.  You can also weight the calculation by another attribute, simply by specifying that as weight.*/
  ///Create a histogram from an attribute
  Graphics_List compute_histogram(string key="", int bins=1, string weight = "");
  //  int compute_histogram(Graphics_List *glist, string key, string weight = "");
  /**< This function has not yet been updated to use the new style ofGraphics_List */
  //@}

  /** @name Graphics Functions
      Functions to create \link Graphics_List graphics objects\endlink or plot data.*/
  //@{
  // Create a Graphics_List from attribute(s)
  Graphics_List create_graphics_list(string s1="",string s2="",string s3="",string s4="",string s5="");
  /**< Each string given will add that attribute to the newly created 
     Graphics_list.  If none are given, the Graphics_List will be entirely
     blank.  Five is a rather arbitrary number, but if you wish to add more
     you can simply use Graphics_List::add_attribute:\n
     \code
     // Assuming we have some list myListPtr from somewhere
     // Make our Graphics_List
     Graphics_List gl = myListPtr->create_graphics_List("attribute1",
                       "attribute2", "attribute3", "attribute4", "attribute5");
     
     // We want a sixth attribute... no problem
     gl.add_attribute(myListPtr, "attribute6");
     \endcode
  */
  // Plot, and return an associated graphics list
  Graphics_List plot(string s1="",string s2="",string s3=""); 
  /**< The Graphics_List returned will contain up to the three attributes
     given to the plot call.  To add additional attributes, see the example in
     List::create_graphics_list.  This function is good for getting a quick
     look at some aspects of the data.
  */
  //@}

  /** @name Utilities
      Currently only parallel sorting and display, but more may follow.*/
  //@{
  // Represents the List.
  // This becomes the Python __repr__ function.
  string represent(){
    stringstream ss;
    ss << "Attributes (#elements = " << nElements << "):\n";
    map<string, Attribute>::iterator iter;
    for(iter = attributes.begin(); iter != attributes.end(); iter++)
      ss << "  " << iter->first << "\n";
    return ss.str();
  }
  //@}
 private:

  int nElements; ///< Number of elements in each attribute
  string posKey; ///< Which attribute is the position one?
  list<int> selection; ///< Current selection
  vector<string> selectionHistory; ///< Past selection information
};

/** \class Grid Data.h "Data.h"
    \brief Grid is a specific Data which is n-dimensional.
    
    It is specified by the number of grid cells, the limits (in some rectangular sense), and the Geometry on which those limits are imposed.*/

class Grid: public Data{
 public:

  /** @name Grid Constructors
      Functions to construct Grid objects, must be called by ALL nodes.*/
  //@{
  //! Empty Grid Constructor
  Grid();
  //! A better constructor
  Grid(string filename,Region* parent);

  /// Standard Grid constructor, specify a parent region and the grid vector.
  Grid(Region* parent, const vector<int>& nxNew);

  /// Grid constructor with non-conforming Geometry
  /** Grids may also be constructed such that they have a different
   * Geometry from their parent region or particle.  For example , they
   * may be a cartesian image of a galaxy which is, itself, within a
   * spherical geometry...  In this case, we'll want to construct from
   * Geometry...
   */
  Grid(Geometry* geo, const vector<int>& nxNew);
  //@}
  /** @name Functions to Add Attributes
      Functions to add attributes and set the value of them.  They return 0 on sucess and -1 on a failure.  ALL processors must call these functions to create the attribute.  newData is purely a dummy value to give the created Attribute type information.*/
  
  Grid(const vector<int>& nxNew);

  //@{
  ///Add an attribute

  

  template<class T> int add_attribute(string key,T* newData) 
    {
      attributes[key]=Attribute(nx,newData[0],key,&(this->DataContext));
      return 0;
    }
  /// Other version of add_attribute
  template<class T> int add_attribute(string key,T newData)
    {
      attributes.insert(map<string,Attribute>::value_type(key,Attribute(nx,newData,key,&(this->DataContext))));
      return 0;
    }
  
  //! Add a whole Attribute
  int add_attribute(string key, Attribute& newVect) 
    {
      newVect.set_name(key);

    /* 
       This condition is supposed to check whether the Grid already has its own 
       non-default Context. If not the Grid is going to take it from the first 
       Attribute added to it if the Attribute has its own Context.
       If the Attribute has not its own Context, the Grid default Context is kept 
       until the first Attribute with a Context is added.
    */
      if (this->DataContext.get_n_grid_points().size()==1 && 
	  this->DataContext.get_n_grid_points()[0]==2 && 
	  newVect.get_context_pointer()->get_n_grid_points().size()!=1)
	{
	  //getting the Context from the Attribute
	  this->set_context(*(newVect.get_context_pointer()));
	}
      else
	{
	  //providing the Attribute with a Context
	  newVect.set_context(&(this->DataContext));
	}
      
      attributes.insert(map<string,Attribute>::value_type(key,newVect));
      return 0;
    }

  //@}
  
   /** @name Function to Add Data
      set_attribute puts the actual data into the attribute.
      Warning: There is no safety checks to prevent collisions.
      Be careful to not have more than one process try and set_attribute to the same range of of the attribute at the same time.
  */
  //@{
  /// Put data to the entire range of the attribute
  template<class T> int set_attribute(string key, T* newData) {
    if(attributes.find(key)!=attributes.end()){
      int nTot=nx[0];
      for (int i=1; i < myGeometry.get_ndims(); i++)
	nTot*=nx[i];
      return attributes[key].put(0,nTot-1,newData);
    }
    else
       return -1;
  }

  /// This set attribute takes lo and hi positions to write to a range
  template<class T> int set_attribute(string key, int iLo, int iHi, T* newData){
    if(attributes.find(key)!=attributes.end())
      return attributes[key].put(iLo,iHi,newData);
    else
      return -1;
  }
  /// Can also pass it a vector
  template<class T> int set_attribute(string key, int iLo, int iHi, vector<T>*& newData){
    if(attributes.find(key)!=attributes.end())
      return attributes[key].put(iLo,iHi,&newData[0]);
    else
      return -1;
  }

  /// This function fills an entire attribute with some constant value.  
  /// Good for say zeroing it out
  template<class T> int fill_attribute(string key, T tConst)
    {
      if(attributes.find(key)!=attributes.end())
	return attributes[key].fill(tConst);
      else
	return -1;
    }

  /// Set one specific element to the given value
  template<class T> int set_element_attribute(string key, T tValue, int idx)
    {
      if(attributes.find(key)!=attributes.end())
	return attributes[key].set_element(idx,tValue);
      else
	return -1;
    }
  //@}
  
  /** @name Position/Index Functions
      Functions to convert index<->position.*/
  //@{
  /// calculate_index(vector) returns the scalar position based on the ordered
  /// n-tuplet input
  int calculate_index(vector<int> x0);
  
  /// Given a scalar index on the grid, returns the coordinates.
  vector<int> calculate_coords(int idx);
  //@}

  /** @name Computations
      Compute useful things from the Grid.*/
  //@{
  
  // Takes the FFT of some grid attribute.  Arguments are the name of
  // the FFT density field to be created (assumed, for now, to be 
  // complex) and the attribute key of the real-space grid.  In
  // principle, this can include other constraints (real-real, etc.)
  Grid* FFT(string outGridTag,string inGridTag);

  // Return a power spectrum
  Graphics_Grid power_spectrum(string outGridTag,string inGridTag);
  /**< Arguments are the output Grid tag and the attribute key of the
     real-space data.*/
  //@}

  /** @name Graphics Functions
      Functions to create \link Graphics_Grid graphics objects\endlink or plot data.*/

  //@{
  // Plot the grid attribute and return a graphics grid
  Graphics_Grid plot(string s1="",int nDims=-1);
  /**< The default value of nDims (-1) will tell it to attempt to plot in the
      highest dimensionality possible, which depends upon the Grid as well as
      the plotting routines available.  You can, however, specify nDims and
      (so long as it is possible) see that dimensionality plot specifically.
  */

  // Create a Graphics_Grid containing up to 5 attributes
  Graphics_Grid create_graphics_grid(string s1="",string s2="",string s3="",string s4="",string s5="");
  /**< See List::create_graphics_list for an example of how to add additional
     attributes if needed.  These two functions are essentially identical.*/

  //@}


  /** @name Utilities
      Functions that don't currently have a real home group.*/
  //@{
  // Accessor function for nx
  vector<int> get_nx() { return nx;}

 // Represents the Grid
  string represent(){
    stringstream ss;
    ss << "Dimensionality: (";
    for(size_t i = 0; i < nx.size(); i++) {
      ss << nx[i];
      if(i < nx.size()-1)
	ss << " , ";
    }
    ss << ")\n";
    ss << "Attributes: \n";
    map<string, Attribute>::iterator iter;
    for(iter = attributes.begin(); iter != attributes.end(); iter++)
      ss << "  " << iter->first << "\n";
    return ss.str();
  }
  //@}


 private:
  //nx is now a data member of Data not needed here anymore
  //  vector<int> nx; ///< Grid cell vector
  vector<int> offset; ///< Vector for offset information
};



#endif



