/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** \file Vector2.h
    \brief Contains Vector2, a 2D vector class

        This file includes:                               \n\n

	-# definition of class vec                        \n
	-# definitions of operators                       \n
		- [] (component)                          \n
		- vector unary -                          \n
		- *, ^ (vector dot and cross products)    \n
		- vector +, -                             \n
		- scalar *, /
		- bool operators ==, !=                   \n
		- operators +=, -=, *=, /=                \n
		- I/O operators >>, <<                    \n
	-# functions square, abs, abs1                    \n
*/

//Based on Vector3

#ifndef  BASIN_VECTOR2_H
# define  BASIN_VECTOR2_H

#include "basin_base.h"

//-----------------------------------------------------------------------------
//  Vector2  --  a class for 2-dimensional vectors
//-----------------------------------------------------------------------------

/** \class Vector2 Vector2.h "Vector2.h"
    \brief A class for 2-dimensional vectors.

    This is the 2D equivalent of Vector3, works exactly the same.
*/
class Vector2
{
 private:
  double element[2];
  // friend class boost::serialization::access;
 public:
  /** \name Vector2 Constructors
      Functions to create a Vector2.*/
  //@{
  //! Default: initialize to zero.
  Vector2(double c = 0.0)
  {element[0] = element[1] = c;}
  //! Explicitly set each component
  Vector2(double x, double y)
  {element[0] = x; element[1] = y;}
  //@}
template<class Archive>
void serialize(Archive & ar, const unsigned int version)
  {
    ar & element[0];
    ar & element[1];
  }

  /** @name Vector2 Operators
      Operators defined on the Vector2 class.*/
  //@{
  //! [] component operator
  /*! []: the return type is declared as a reference (&), so that it can
      be used on the left-hand side of an assignment, as well as on the
      right-hand side, i.e.  v[1] = 3.14  and  x = v[2]  are both allowed
      and work as expected.*/
  double & operator [] (int i)       {return element[i];}
  
  //! print function for outpout operator
  inline void print() {cout << element[0] << " " << element[1] << " "
				  << "\n";}
  //! Unary -
  Vector2 operator - ()
    {return Vector2(-element[0], -element[1]);}

  //! Dot product.
  double operator * (const Vector2& b)
	    {return element[0]*b.element[0]
		  + element[1]*b.element[1];}

  //! Outer product.
  Vector2 operator ^ (const Vector2 &b)
    {return Vector2(0.0,0.0);}
  
  //! Vector2 +
  Vector2 operator + (const Vector2 &b)
    {return Vector2(element[0]+b.element[0],
		    element[1]+b.element[1]);}
  //! Vector2 -
  Vector2 operator - (const Vector2 &b)
    {return Vector2(element[0]-b.element[0],
		    element[1]-b.element[1]);}
  //@}

  //! Scaler +
  friend Vector2 operator + (double, const Vector2 & );
  //! Other way
  friend Vector2 operator + (const Vector2 &, double);
  
  //! Scalar *
  friend Vector2 operator * (double, const Vector2 & );
  //! Other way
  friend Vector2 operator * (const Vector2 &, double);
  //! Scalar /
  friend Vector2 operator / (const Vector2 &, double);
  //! Bool ==
  /*! Bool ==, !=: apparently can't do bit-by-bit compare because the
     member data are private.  Hence these need to be friends in order
     to apply these operators.  Almost never used, however...*/
  friend bool operator == (const Vector2 &, const Vector2 &);
  //! Bool !=
  /*! Bool ==, !=: apparently can't do bit-by-bit compare because the
     member data are private.  Hence these need to be friends in order
     to apply these operators.  Almost never used, however...*/
  friend bool operator != (const Vector2 &, const Vector2 &);

  /** @name Vector2 Operators
      Continued...*/
  //@{
  // Post operations (again...) -- the GCC3 december 2002 sage:
  // this is a some interesting way to resolve the ambiguity of
  // resolving  v*s vs. s*v and the confusion of having a v*v with
  // a non-explicit Vector2 constructor from a scalar.....

  //! Post operator *
  Vector2 operator * (const double b)
    {return Vector2(element[0]*b, element[1]*b); }
  
  //! Post operator +
  Vector2 operator + (const double b)
    {return Vector2(element[0]+b, element[1]+b); }
  
  //! Vector2 +=
  Vector2& operator += (const Vector2& b)
    {element[0] += b.element[0];       
    element[1] += b.element[1];
    return *this;}
  
  //! Vector2 -=
  Vector2& operator -= (const Vector2& b)
    {element[0] -= b.element[0];
    element[1] -= b.element[1];
    return *this;}
  
  //! Vector2 *=
  Vector2& operator *= (const double b)
    {element[0] *= b;
    element[1] *= b;
    return *this;}
  
  //! Vector2 /=
  Vector2& operator /= (const double b)
    {element[0] /= b;
    element[1] /= b;
    return *this;}
  //@}
  /** @name I/O Operators
      Stream operators for Vector2 I/O.*/
  //! Output stream operator
  friend ostream & operator << (ostream & , const Vector2 & );
  //! Input stream operator
  friend istream & operator >> (istream & , Vector2 & );
};

inline  ostream & operator << (ostream & s, const Vector2 & v)
	    {return s << v.element[0] << "  " << v.element[1];}

inline  istream & operator >> (istream & s, Vector2 & v)
	    {s >> v.element[0] >> v.element[1];
	     return s;}

inline  double square(Vector2 v) {return v*v;}
inline  double abs(Vector2 v)    {return sqrt(v*v);}

//! Another measure of Vector2 magnitude; less work than abs():

inline  double abs1(Vector2 v)   {return fabs(v[0]) + fabs(v[1]);}

inline  Vector2 operator + (double b, const Vector2 & v)
	    {return Vector2(b+v.element[0],
			b+v.element[1]);}

inline  Vector2 operator + (const Vector2 & v, double b)
	    {return Vector2(b+v.element[0],
			b+v.element[1]);}

inline  Vector2 operator * (double b, const Vector2 & v)
	    {return Vector2(b*v.element[0],
			b*v.element[1]);}

inline  Vector2 operator * (const Vector2 & v, double b)
	    {return Vector2(b*v.element[0],
			b*v.element[1]);}

inline  Vector2 operator / (const Vector2 & v, double b)
	    {return Vector2(v.element[0]/b,
			v.element[1]/b);}

inline  bool operator == (const Vector2& u, const Vector2& v)
	    {return (u.element[0] == v.element[0]
		     && u.element[1] == v.element[1]);}

inline  bool operator != (const Vector2& u, const Vector2& v)
	    {return !(u==v);}

#endif
