/*
 *  
 *  Copyright (C) 2003  Starlab team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** \file Vector3.h
    \brief Contains Vector3, a 3D vector class

        This file includes:                               \n\n

	-# definition of class vec                        \n
	-# definitions of operators                       \n
		- [] (component)                          \n
		- vector unary -                          \n
		- *, ^ (vector dot and cross products)    \n
		- vector +, -                             \n
		- scalar *, /
		- bool operators ==, !=                   \n
		- operators +=, -=, *=, /=                \n
		- I/O operators >>, <<                    \n
	-# functions square, abs, abs1                    \n
*/

       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\ ~
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\ ~
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\ ~
 //                                                       //            _\|/_
//=======================================================//              /|\ ~

//  Vector3.h: 3D vec operations include file
//.............................................................................
//    version 1:  Dec 1992   Piet Hut, Steve McMillan, Jun Makino
//    version 2:  Jun 2003   Steve McMillan: vector now vec to avoid
//                           conflict with STL
//    version 3:  Apr 2005   BASIN: work arounds to comply with BasinArray
//.............................................................................
//     This file includes:
//
//	1) definition of class Vector3
//	2) definitions of operators
//		[] (component)
//		vector unary -
//		*, ^ (vector dot and cross products)
//		vector +, -
//		scalar *, /
//		bool operators ==, !=
//		operators +=, -=, *=, /=
//		I/O operators >>, <<
//	3) functions square, abs, abs1
//.............................................................................

#ifndef  STARLAB_VECTOR_H
# define  STARLAB_VECTOR_H

#include "basin_base.h"

//-----------------------------------------------------------------------------
//  Vector3  --  a class for 3-dimensional vectors
//-----------------------------------------------------------------------------

/** \class Vector3 Vector3.h "Vector3.h"
    \brief A class for 3-dimensional vectors.

    The Vector3 class comes from Starlab, some minor adjustments have been made (mostly formatting to be more friendly for documentation).
*/
class Vector3
{
 private:
  double element[3];
  // friend class boost::serialization::access;
 public:
  /** \name Vector3 Constructors
      Functions to create a Vector3.*/
  //@{
  //! Default: initialize to zero.
  Vector3(double c = 0.0)
  {element[0] = element[1] = element[2] = c;}
  //! Explicitly set each component
  Vector3(double x, double y, double z)
  {element[0] = x; element[1] = y; element[2] = z;}
  //@}
template<class Archive>
void serialize(Archive & ar, const unsigned int version)
  {
    ar & element[0];
    ar & element[1];
    ar & element[2];
  }

  /** @name Vector3 Operators
      Operators defined on the Vector3 class.*/
  //@{
  //! [] component operator
  /*! []: the return type is declared as a reference (&), so that it can
      be used on the left-hand side of an assignment, as well as on the
      right-hand side, i.e.  v[1] = 3.14  and  x = v[2]  are both allowed
      and work as expected.*/
  double & operator [] (int i)       {return element[i];}
  
  //! print function for outpout operator
  inline void print() {cout << element[0] << " " << element[1] << " "
				  << element[2] << "\n";}
  //! Unary -
  Vector3 operator - ()
    {return Vector3(-element[0], -element[1], -element[2]);}

  //! Dot product.
  double operator * (const Vector3& b)
	    {return element[0]*b.element[0]
		  + element[1]*b.element[1]
		  + element[2]*b.element[2];}

  //! Outer product.
  Vector3 operator ^ (const Vector3 &b)
    {return Vector3(element[1]*b.element[2] - element[2]*b.element[1],
		    element[2]*b.element[0] - element[0]*b.element[2],
		    element[0]*b.element[1] - element[1]*b.element[0]);}
  
  //! Vector3 +
  Vector3 operator + (const Vector3 &b)
    {return Vector3(element[0]+b.element[0],
		    element[1]+b.element[1],
		    element[2]+b.element[2]);}
  //! Vector3 -
  Vector3 operator - (const Vector3 &b)
    {return Vector3(element[0]-b.element[0],
		    element[1]-b.element[1],
		    element[2]-b.element[2]);}
  //@}

  //! Scaler +
  friend Vector3 operator + (double, const Vector3 & );
  //! Other way
  friend Vector3 operator + (const Vector3 &, double);
  
  //! Scalar *
  friend Vector3 operator * (double, const Vector3 & );
  //! Other way
  friend Vector3 operator * (const Vector3 &, double);
  //! Scalar /
  friend Vector3 operator / (const Vector3 &, double);



  //! Scaler +
  friend Vector3 operator + (float, const Vector3 & );
  //! Other way
  friend Vector3 operator + (const Vector3 &, float);
  
  //! Scalar *
  friend Vector3 operator * (float, const Vector3 & );
  //! Other way
  friend Vector3 operator * (const Vector3 &, float);
  //! Scalar /
  friend Vector3 operator / (const Vector3 &, float);
  //! Bool ==
  /*! Bool ==, !=: apparently can't do bit-by-bit compare because the
     member data are private.  Hence these need to be friends in order
     to apply these operators.  Almost never used, however...*/
  friend bool operator == (const Vector3 &, const Vector3 &);
  //! Bool !=
  /*! Bool ==, !=: apparently can't do bit-by-bit compare because the
     member data are private.  Hence these need to be friends in order
     to apply these operators.  Almost never used, however...*/
  friend bool operator != (const Vector3 &, const Vector3 &);

  /** @name Vector3 Operators
      Continued...*/
  //@{
  // Post operations (again...) -- the GCC3 december 2002 sage:
  // this is a some interesting way to resolve the ambiguity of
  // resolving  v*s vs. s*v and the confusion of having a v*v with
  // a non-explicit Vector3 constructor from a scalar.....

  //! Post operator *
  Vector3 operator * (const double b)
    {return Vector3(element[0]*b, element[1]*b, element[2]*b); }
  
  //! Post operator +
  Vector3 operator + (const double b)
    {return Vector3(element[0]+b, element[1]+b, element[2]+b); }
  
  //! Vector3 +=
  Vector3& operator += (const Vector3& b)
    {element[0] += b.element[0];       
    element[1] += b.element[1];
    element[2] += b.element[2];
    return *this;}
  
  //! Vector3 -=
  Vector3& operator -= (const Vector3& b)
    {element[0] -= b.element[0];
    element[1] -= b.element[1];
    element[2] -= b.element[2];
    return *this;}
  
  //! Vector3 *=
  Vector3& operator *= (const double b)
    {element[0] *= b;
    element[1] *= b;
    element[2] *= b;
    return *this;}
  
  //! Vector3 /=
  Vector3& operator /= (const double b)
    {element[0] /= b;
    element[1] /= b;
    element[2] /= b;
    return *this;}
  //@}
  /** @name I/O Operators
      Stream operators for Vector3 I/O.*/
  //! Output stream operator
  friend ostream & operator << (ostream & , const Vector3 & );
  //! Input stream operator
  friend istream & operator >> (istream & , Vector3 & );
};

inline  ostream & operator << (ostream & s, const Vector3 & v)
	    {return s << v.element[0] << "  " << v.element[1]
		      << "  " << v.element[2];}

inline  istream & operator >> (istream & s, Vector3 & v)
	    {s >> v.element[0] >> v.element[1] >> v.element[2];
	     return s;}

inline  double square(Vector3 v) {return v*v;}
inline  double abs(Vector3 v)    {return sqrt(v*v);}

//! Another measure of Vector3 magnitude; less work than abs():

inline  double abs1(Vector3 v)   {return fabs(v[0]) + fabs(v[1]) + fabs(v[2]);}

inline  Vector3 operator + (double b, const Vector3 & v)
	    {return Vector3(b+v.element[0],
			b+v.element[1],
			b+v.element[2]);}

inline  Vector3 operator + (const Vector3 & v, double b)
	    {return Vector3(b+v.element[0],
			b+v.element[1],
			b+v.element[2]);}

inline  Vector3 operator * (double b, const Vector3 & v)
	    {return Vector3(b*v.element[0],
			b*v.element[1],
			b*v.element[2]);}

inline  Vector3 operator * (const Vector3 & v, double b)
	    {return Vector3(b*v.element[0],
			b*v.element[1],
			b*v.element[2]);}

inline  Vector3 operator / (const Vector3 & v, double b)
	    {return Vector3(v.element[0]/b,
			v.element[1]/b,
			v.element[2]/b);}

inline  Vector3 operator / (const Vector3 & v, float b)
	    {return Vector3(v.element[0]/b,
			v.element[1]/b,
			v.element[2]/b);}


inline  bool operator == (const Vector3& u, const Vector3& v)
	    {return (u.element[0] == v.element[0]
		     && u.element[1] == v.element[1]
		     && u.element[2] == v.element[2]);}

inline  bool operator != (const Vector3& u, const Vector3& v)
	    {return !(u==v);}

#endif
