/*
 *  Simple C++ gnuplot API for use with StarCluster tools.
 *  Copyright (C) 2003  StarCluster team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef GNUPLOT_HH
#define GNUPLOT_HH

#include <string>
#include <iostream>
#include <vector>
#include <stdio.h>
using namespace std;


//! \a Gnuplot API
/*!
 * This class implements a somewhat incomplete but very convenient and simple to
 * use \a Gnuplot API for the \a StarCluster package.
 */
class gnuplot {

public:
  //! if set, autoscale mode is used for every plot
  /*!
   * If set, autoscale mode is used for every plot, not just the first one.
   * Otherwise, only the first plot is auto-scaled, and the following ones
   * inherit those axis limits. This is off by default as it produces very
   * annoying rescaling and recentering of plots when only the central portion
   * is "interesting."
   * \warning The action of this variable isn't completely reversible. More
   * precisely, the limits are saved only from the first dataset and if
   * autoscaling is desired after the first dataset was plotted, \c
   * set_xrange("*:*")/set_yrange("*:*") must be used instead. In other words,
   * \c autoscale has the described effect only before the first call to
   * flush(), where it's decided whether or not to save the limits.
   * \bug When using the Gnuplot feature that saves and restores plot limits,
   * the first plot may have different axis labels and tics than all successive
   * ones. To avoid this, either #autoscale or set_xrange()/set_yrange() must be
   * used.
   * \bug This mode is incompatible with tiling enabled by
   * set_multiplot(unsigned char, unsigned char) as \a Gnuplot only clears the
   * area defined by its variables, \a origin and \a size, and thus doesn't
   * clear the axes as well as their labels and tics.
   */
  bool autoscale;
  //! number of seconds to pause between plots
  /*!
   * When this variable is set to a positive value, it's interpreted as the
   * number of seconds to pause between successive plots (or screens, if in
   * multiplot mode). If it's equal to -1, \a StarCluster will exit. And if the
   * value of \c pause is less than -1, \a StarCluster will wait for a key press
   * before producing the next plot/screen unless \c stdin isn't connected to a
   * terminal, e.g., data is being read from \c stdin. This value can be ignored
   * under two other circumstances: flush(false) was called (instead of flush())
   * or no actual \a Gnuplot session is active, instead scripts or data are
   * being output as a result of set_fp().
   */
  int pause;
  //! make 3-D plots
  /*!
   *  make 3-D plots
   */
  bool plot3d;
  //! style of the plot, i.e., \c points, \c lines, \c dots, ...
  /*!
   * This determines the style of the plot. Any valid \a Gnuplot style can be
   * specified: \c points, \c lines, \c linespoints, ...
   */
  string style;
  //! time associated with the current plot
  /*!
   * Holds the time associated with the current plot. If set, this number will
   * be used to add time information to the title of the plot. To make a simple
   * X-Y projection of an N-body system: \code
   * gnuplot gp;
   * gp.title = "X-Y Positions";
   * gp.xlabel = "x", gp.ylabel = "y";
   * while (dyn* r = get_next_dyn()) {
   *   gp.time = r->get_system_time();
   *   for_all_leaves(dyn, r, d)
   *     gp << d->get_pos()[0] << d->get_pos()[1];
   *   gp << flush;
   *   rmtree(r);
   * } \endcode
   * If it's desirable, on the other hand, to omit this information, \c time
   * must be set to \c INVALID_TIME.
   * \see set_title()
   */
  float time;
  //! title of the plot
  /*!
   * This variable holds the title of the plot and is used before the first
   * dataset is given with operator<<(double). It's commonly used in tools
   * utilizing the \c gnuplot library to set the default title before parsing
   * command-line options. Using set_title() would be equivalent except for one
   * subtle difference: the \a Gnuplot code would be output immediately and if
   * the output were changed with set_fp() using comand-line options, which are
   * parsed after setting the default, the code setting the title would never be
   * output to the specified stream. When using the variable, however,
   * set_title() doesn't have to be called explicitly and is only necessary if
   * the title is to be changed after the first dataset was plotted: \code
   * gnuplot gp;
   * gp.title = "Title Example";        // convenient to use the variable here
   * gp << 1 << 2 << ... << flush;
   * gp.set_title("Another Example");   // here, the function form must be used
   * gp << 2 << 3 << ... << flush; \endcode
   * The same difference applies to the #xlabel/set_xlabel() and
   * #ylabel/set_ylabel() pairs as well as, partly, to #autoscale.
   */
  string title;
  //! label for the x axis
  /*!
   * Label for the X axis. \see #title
   */
  string xlabel;
  //! label for the y axis
  /*!
   * Label for the Y axis. \see xlabel
   */
  string ylabel;
  //! label for the z axis
  /*!
   * Label for the Z axis. \see xlabel
   */
  string zlabel;
  //! constructor for the \c gnuplot class
  /*!
   * Creates a default \c gnuplot object and sets up a pipe to a \a Gnuplot
   * process with \c popen().
   */
  gnuplot(const bool = true);
  //! destructor for the \c gnuplot class
  /*!
   * Closes the pipe to the object's \a Gnuplot process with \c pclose() and
   * deallocates its memory.
   */
  ~gnuplot();
  //! ends the current plot and flushes all data to \a Gnuplot
  /*!
   * This function is used to terminate the current plot and flush the data to
   * the \a Gnuplot process. It's optional only in the simplest cases when it's
   * sufficient for ~gnuplot(), called explicitly or at the end of scope, to
   * complete the plot: \code
   * #include "gnuplot.hh"
   *
   * int main(void) {
   *   gnuplot gp;
   *   gp << 1 << 2 << ...;
   * } \endcode
   * In situations where more sophisticated features are used, e.g., tiled
   * plots, multiple graphs in one plot, or animations, flush() must be invoked
   * to signal the end of datasets. It can be called in two ways, explicitly:
   * \code gp.flush(); \endcode or, more naturally, through the overloaded
   * <tt>operator<<(gnuplot& (*f)(gnuplot&))</tt> and a nonmember
   * <tt>gnuplot& flush(gnuplot&)</tt>: \code gp << flush; \endcode
   * \note The optional argument is intended primarily for internal use. When
   * flush(false) is called after a dataset, the value of #pause is ignored. For
   * example: \code
   * gp.pause = -1;
   * gp << 1 << 2 << ... << 3.14159, gp.flush(false);   // no pause
   * gp << 2 << 3 << ... << flush;                      // pause here \endcode
   * Notice that \code gp << flush(false); \endcode syntax isn't supported; it's
   * another indication that this usage is deprecated.
   */
  gnuplot& flush(const bool = true);
  //! passes a string to \a Gnuplot as a command
  /*!
   * Used to pass arbitrary commands to Gnuplot. This member function is
   * intended to extend \a Gnuplot support beyond the provided interface. For
   * example, if \c gp is a \c gnuplot object, \code
   * gp << "set title \"{/Symbol=18 r}(t)\"\n"; \endcode will set the title of
   * the plot to rho(t) \f$(\rho(t))\f$ in 18 pt \a Symbol font. Many provided
   * interface member functions are implemented like this. Another major use of
   * this function is to plot functions, perhaps together with data: \code
   * gnuplot gp;
   * gp.set_multiplot();
   * gp << 0.0 << 0.0 << 1.57 << 1 << 3.14 << 0.0 << flush;
   * gp << "plot sin(x) with lines 2\n"; \endcode
   * This will plot the three points and a \a sine curve through them. Notice
   * that the data is given first - this is the preferred order as it defines
   * the limits of the plot, and the function is plotted only in that area.
   * \note When this or any other function that issues commands is used
   * during an active plot, i.e., after operator<<(double) but before flush(),
   * the current plot is terminated with flush(). This is, of course, because
   * \a Gnuplot will, otherwise, produce a syntax error as it will not interpret
   * commands in the middle of reading numeric data.
   */
  gnuplot& operator<<(const string);
  //! passes a number to \a Gnuplot for plotting
  /*!
   * This is the most often used function of this class, and the one that, by
   * far, contains the most logic. It's used to send numbers to be plotted to \a
   * Gnuplot. For example, if \c gp is a \c gnuplot object,
   * \code gp << 2 << 3; \endcode or \code gp << 2; gp << 3; \endcode will send
   * the coordinate pair, (2,3), to Gnuplot. It will not, however, be plotted
   * right away, as \c gp will be expecting more numbers, unless it's the last
   * statement in the program (or gp.~gnuplot() is invoked through a different
   * mechanism) or gp.flush() is called explicitly or, equivalently, by
   * \code gp << flush; \endcode The simplicity of this entire interface comes
   * at the cost of the complexity of this function: it contains about as much
   * code as the rest of the class combined. The goal was such that after
   * setting some flags describing the desired plot configuration, the user
   * would simply send numbers without other formating information and the \c
   * gnuplot object would keep track of the current state. For example, to plot
   * three datasets in one graph: \code
   * gnuplot gp;
   * gp.set_graphs(3);
   * gp << 1 << 2;       // first point of the first dataset
   * gp << 3 << 4;       // first point of the second dataset
   * gp << 5 << 6;       // first point of the third dataset
   * gp << 7 << 8;       // second point of the first dataset
   * ...                 // same pattern continues
   * gp << flush;        // plot is finished \endcode
   * This particular sequence is very convenient as many \a Starlab functions,
   * e.g., compute_general_mass_radii(), produce output in exactly this order.
   */
  gnuplot& operator<<(const double);
  gnuplot& operator<<(gnuplot& (*f)(gnuplot&));
  //! resets output back to \a Gnuplot
  /*!
   * Used to reverse the action of set_fp(). After invoking this member
   * function, the pipe is restored to point to the original \a Gnuplot process.
   */
  void reset_fp();
  //! redirects output away from the \a Gnuplot pipe and to a file
  /*!
   * This member function is intended for debugging purposes or if tweaking the
   * produced \a Gnuplot script is desired. It redirects the output of the API
   * to a file and away from the \a Gnuplot process. The file pointer is usually
   * provided by c_file_ptr() from \c tool.hh; it must be valid and point to a
   * file opened for writing. If the second, optional, argument is \c true,
   * then, instead of the entire script (commands and data), only the data is
   * output. It is, however, in the same order as would have been output to \a
   * Gnuplot, which is not necessarily the same as input order if plotting more
   * than one dataset with set_graphs(). For example: \code
   * gnuplot gp;
   * gp.set_fp(stdout, true);
   * gp.set_graphs(3);
   * gp << 1 << 2;
   * gp << 2 << 3;
   * gp << 3 << 4;
   * ... \endcode
   * will produce this output: \code
   * 1 2
   * ...
   * <newline>
   * 2 3
   * ...
   * <newline>
   * 3 4
   * ...
   * <newline> \endcode
   * It's action can be reversed with reset_fp(), which resets the file pointer
   * back to the original \a Gnuplot process.
   */
  void set_fp(FILE* const, const bool = false);
  //! creates multiple graphs in one plot
  /*!
   * This member function allows to plot multiple datasets together. The data is
   * expected in the order described under operator<<(double) and set_fp().
   * \see set_multiplot(unsigned char, unsigned char) if a "dataset at a time"
   * order is preferred
   */
  void set_graphs(unsigned short);
  //! enables log scaling of specified axes
  /*!
   * Enable a logarithmic scale for the specified axes. The argument can specify
   * the desired axes as well the the log base. If no argument is supplied, the
   * defaults, all axes and base 10, are used. Some examples are: \code
   * gp.set_log();             // log base 10 scaling of all axes
   * gp.set_log("x");          // log base 10 scaling of the X axis
   * gp.set_log("xy 2.71828")  // natural log scaling of both axes \endcode
   */
  void set_log(const string = "");
  //! creates tiled plots or plots of multiple datasets
  /*!
   * Creates tiled plots or plots of multiple datasets. If called without
   * arguments, it simply plots all datasets in the same window until
   * set_nomultiplot() is invoked. Otherwise, it creates a tile of plots of the
   * specified, M by N, configuration. In the first case, it differs from
   * set_graphs() in two respects: the expected order of data, which a dataset
   * at a time in this mode, and the inability to specify in advance the number
   * of desired plots. In particular, both functionalities can't be utilized
   * simultaneously, i.e., if several datasets are to be be plotted and the
   * plots tiled, set_graphs() must also be used. For example, to produce a 3x3
   * tile of plots: \code
   * gnuplot gp;
   * gp.set_multiplot(3,3);          // or gp.set_multiplot("3")
   * gp << 1 << 2 << ... << flush;   // the first plot
   * gp << 3 << 4 << ... << flush;   // the second one, placed next to it
   * ...                             // and so on, for seven more plots \endcode
   * To plot several datasets together in the same plot: \code
   * gnuplot gp;
   * gp.set_multiplot();
   * gp << 1 << 2 << ... << flush;   // the first dataset
   * gp << 3 << 4 << ... << flush;   // the second one, placed in the same plot
   * ...                             // and so on, ... \endcode
   * \bug It's important to note that in order to produce good-looking tiled
   * plots of (nearly) arbitrary size, the algorithm for calculating spaces and
   * placements would have to be greatly improved - it's already fairly
   * complex! As a result, only plots of "reasonable" size (M,N < 5) can be
   * expected to look acceptable. This issue is further complicated by the
   * variety of available fonts as well as \a Gnuplot's handling of various
   * "terminals", or output devices.
   * \see
   *   \li set_graphs() if a different input order is preferred
   *   \li #autoscale for incompatible usage
   */
  void set_multiplot(const unsigned char = 1, const unsigned char = 1);
  //! creates tiled plots or plots of multiple datasets
  /*!
   * This function parses its string argument and calls
   * set_multiplot(unsigned char, unsigned char) with appropriate
   * arguments. It's intended to be given a command-line option directly, which
   * must be in one of two \c scanf() formats:
   * \li \b MxN (\c "%hhux%hhu") - M by N tile of plots
   * \li \b N (\c "%hhu") - N by N square tile of plots
   * \see set_multiplot(unsigned char, unsigned char)
   */
  void set_multiplot(const string);
  //! disables log scaling of specified axes
  /*!
   * Disable logarithmic scaling for the specified axes. For this function, the
   * argument must either specify the axis or be empty, in which case linear
   * scaling is restored for all axes.
   */
  void set_nolog(const string = "");
 //! disables multiplot mode
  /*!
   * Disables multiplot mode previously enabled with
   * set_multiplot(unsigned char, unsigned char).
   * \see set_graphs()
   */
  void set_nomultiplot();
  //! creates a \a Postscript plot
  /*!
   * Instead of producing an \a X display plot, make a \a Postscript file. If
   * the argument is \c -, it will be sent to \a stdout, otherwise to the
   * requested file.
   */
  void set_output(const string);
  //! sets the title of the plot
  /*!
   * Sets the title of the plot; it will be placed at the top of the plot in the
   * heading. And if #time is not \c INVALID_TIME, it will also be followed, on
   * the next line, by \c "t = X", where X is the time associated with the
   * snapshot. Additionally, when producing \a Postscript plots, it may be
   * desirable to use \a LaTeX escapes for Greek letters and other symbols for
   * titles as well as axis labels. These are defined in \a Gnuplot
   * documentation but some examples are:
   * \li \c rho(t): \c "{/Symbol r}(t)" \f$\Longrightarrow\rho(t)\f$
   * \li \c erf(x): \c "2/{/Symbol \326 p} {{/Symbol=18 \362}@_0^x} e^{-u^2}du"
   *     \f$\Longrightarrow2/\sqrt\pi\int_0^x e^{-u^2}du\f$
   * \see #title
   */
  void set_title(const string);
  //! sets the label of the X axis
  /*!
   * Sets the label of the X axis.
   * \see
   *   \li set_title() for examples of \a LaTeX escapes
   *   \li #xlabel for slightly different semantics
   */
  void set_xlabel(const string);
  //! sets the extend of the X axis
  /*!
   * The range of the X axis is passed as a string in \a Gnuplot format,
   * omitting the outer brackets. For example, to set the extend of the X axis
   * from -2 to an autoscaled value, \code gp.set_xrange("-2:*") \endcode
   */
  void set_xrange(const string);
  //! sets the label of the Y axis
  /*!
   * Sets the label of the Y axis. \see set_xlabel()
   */
  void set_ylabel(const string);
  //! sets the extend of the Y axis
  /*!
   * Sets the limits of the Y axis. \see set_xrange()
   */
  void set_yrange(const string);
  //! sets the label of the Z axis
  /*!
   * Sets the label of the Z axis. \see set_xlabel()
   */
  void set_zlabel(const string);
  //! sets the extend of the Z axis
  /*!
   * Sets the limits of the Z axis. \see set_xrange()
   */
  void set_zrange(const string);
  //! Insert a newline character, used as fix for maps in pm3d
  void insert_nl();
private:
  FILE *fp, *fp_;
  unsigned char n, plots[2];
  unsigned short graphs, graph, plot;
  bool data_only, first_dataset, first_flush, first_number, multiplot, new_plot;
  vector<vector<double> > buf;
  vector<double> first_buf;
  gnuplot(const gnuplot&);
  gnuplot& operator=(const gnuplot&);

};


gnuplot& flush(gnuplot&);

#endif  // GNUPLOT_HH
